from typing import Any

from ..interpreter.abstract import Interpreter


class InterpretationError(Exception):
    def __init__(
        self, message: str, interpreter: Interpreter, state: Any = None  # noqa: ANN401
    ) -> None:
        super().__init__(message)
        self.interpreter = interpreter
        self.state = state or interpreter.state



from malevich_space.schema.schema import SchemaMetadata

from ..collection import Collection
from .base import BaseNode


class CollectionNode(BaseNode):
    collection: Collection
    scheme: str | SchemaMetadata | None = None



from functools import wraps
from typing import Callable, Optional, ParamSpec, TypeVar

import pandas as pd

from .._autoflow.flow import Flow
from .._autoflow.tracer import traced
from .._utility.registry import Registry
from ..models.flow_function import FlowFunction
from ..models.nodes.base import BaseNode
from ..models.nodes.tree import TreeNode
from ..models.task.promised import PromisedTask
from ..models.types import TracedNode, TracedNodes
from .collection import collection

T = TypeVar("T")
R = PromisedTask | TracedNode | TracedNodes
Args = ParamSpec("Args")
reg = Registry()


def flow(
    reverse_id: Optional[str] = None,
    name: Optional[str] = None,
    description: Optional[str] = None,
    dfs_are_collections: bool = False,
) -> Callable[[Callable[Args, T]], FlowFunction[Args, R]]:
    """Creates a flow from a function

    Decorated function will return either a task or traced nodes.
    The task is returned when the function is called from the main context,
    and traced nodes are returned when the function is called from another
    flow. The task can be interpreted by calling `interpret` method with
    any of available interpreters and then run with Malevich API. Traced
    nodes are technical objects that should not be used directly but
    rather passed to other @flow functions or Malevich operators

    Args:
        reverse_id (str, optional): Reverse ID of the flow component. Defaults to the name of the function.
        name (str, optional): Name of the flow. Defaults to None.
        description (str, optional): Description of the flow. Defaults to None.
        dfs_are_collections (bool, optional): Whether to treat pandas.DataFrame as a collection. Defaults to False.

    Returns:
        Callable[[Callable[Args, T]], Callable[Args, T]]: Decorator
    """   # noqa: E501
    def wrapper(function: Callable[Args, T]) -> FlowFunction[Args, R]:
        nonlocal reverse_id, name, description

        reverse_id = reverse_id or function.__name__
        name = name or function.__name__.replace("_", " ").title()
        description = description or (function.__doc__.splitlines()[0] if function.__doc__ and len(function.__doc__.splitlines()) > 0 else "")  # noqa: E501

        @wraps(function)
        def fn(*args: Args.args, **kwargs: Args.kwargs) -> R:
            is_subflow = Flow.isinflow()
            if is_subflow:
                outer_tracer = traced()
                args = list(args)

            with Flow() as _tree:
                __hargs = []
                __hkwargs = {}
                for i, i_arg in enumerate(args):
                    if dfs_are_collections and isinstance(i_arg, pd.DataFrame):
                        args[i] = (
                            i_arg := collection(
                                df=i_arg,
                                name=function.__code__.co_varnames[i],
                            )
                        )
                    if isinstance(i_arg, traced):
                        __hargs.append(traced(BaseNode()))
                    else:
                        __hargs.append(i_arg)

                for k in kwargs:
                    k_arg = kwargs[k]
                    if dfs_are_collections and isinstance(k_arg, pd.DataFrame):
                        kwargs[k] = (
                            k_arg := collection(
                                df=k_arg,
                                name=k
                            )
                        )
                    if isinstance(k_arg, traced):
                        __hkwargs[k] = traced(BaseNode())
                    else:
                        __hkwargs[k] = k_arg
                if is_subflow:
                    __results = function(*__hargs, **__hkwargs)
                else:
                    __results = function(*args, **kwargs)
                t_node = TreeNode(
                    tree=_tree,
                    reverse_id=reverse_id,
                    name=name,
                    description=description,
                )

            if is_subflow:
                outer_tracer.claim(t_node)
                for i, i_arg in enumerate(args):
                    if isinstance(i_arg, traced):
                        # In parent tree
                        bridges = _tree.edges_from(__hargs[i])
                        b_nodes = [(b[1], b[2]) for b in bridges]
                        i_arg._autoflow.calledby(outer_tracer, (i, b_nodes))
                        _tree.prune([__hargs[i]])

                for k in kwargs:
                    k_arg = kwargs[k]
                    if isinstance(k_arg, traced):
                        # New tracer in this tree
                        bridges = _tree.edges_from(__hkwargs[k])
                        b_nodes = [(b[1], b[2]) for b in bridges]
                        k_arg._autoflow.calledby(outer_tracer, (i, b_nodes))
                        _tree.prune([__hkwargs[k]])

                outputs = [
                    traced(
                        TreeNode(
                            **t_node.model_dump(exclude=["underlying_node"]),
                            underlying_node=o.owner,
                        )
                    )
                    for o in (
                        [__results] if isinstance(
                            __results, traced) else __results
                    )
                    if isinstance(o, traced)
                ]

                return outputs[0] if len(outputs) == 1 else outputs
            else:
                return PromisedTask(results=__results, tree=t_node)

        return FlowFunction(fn, reverse_id, name, description)

    return wrapper



from typing import Any, Iterable, Optional

from pydantic import ConfigDict

from ..._autoflow.tracer import traced
from ..._autoflow.tree import ExecutionTree
from .base import BaseNode


class TreeNode(BaseNode):
    tree: ExecutionTree[traced[BaseNode]]
    results: Iterable[traced[BaseNode]] | traced[BaseNode] | None = None
    reverse_id: str
    name: str
    description: str = "Wonderful Flow!"
    underlying_node: Optional[Any] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __eq__(self, __value: object) -> bool:
        return self.uuid == __value.uuid



import json
from typing import Optional
from uuid import uuid4

import pandas as pd
from malevich_space.schema.schema import SchemaMetadata

from .._autoflow import tracer as gn  # engine
from .._utility.schema import pd_to_json_schema
from ..models.collection import Collection
from ..models.nodes import CollectionNode


def collection(
    name: str,
    file: Optional[str] = None,
    df: Optional[pd.DataFrame] = None,
    scheme: SchemaMetadata = None,
    alias: Optional[str] = None,
) -> gn.traced[CollectionNode]:
    """Creates a collection from a file or a dataframe

    The name might be required when working with some of interpreters

    Args:
        name (str, optional): Name of the collection. Defaults to None.
        file (Optional[str], optional): Path to the file. Defaults to None.
        df (Optional[pd.DataFrame], optional): Dataframe. Defaults to None.
        scheme (Optional[SchemaMetadata], optional): Schema of the collection.

    Raises:
        AssertionError: If both file and data are provided

    Returns:
        Collection: an object that is subject to be passed to processors
    """
    assert any(
        [file is not None, df is not None]
    ), "Either file or data must be provided"

    if file:
        assert df is None, "Cannot provide both file and data"
        df = pd.read_csv(file)

    name = name or 'auto-collection-' + uuid4().hex[:8]

    # Retrieve JSON Schema
    # from pandas DataFrame

    pd_scheme = scheme or json.dumps(pd_to_json_schema(df))
    node = CollectionNode(
        collection=Collection(
            collection_id=name,
            collection_data=df,
        ),
        scheme=pd_scheme,
    )

    if alias:
        node.alias = alias

    return gn.traced(node)



from typing import Callable, Generic, ParamSpec, TypeVar

Params = ParamSpec("Params")
R = TypeVar("R")

class FlowFunction(Generic[Params, R]):
    def __init__(
        self,
        f: Callable[Params, R],
        reverse_id: str,
        name: str,
        description: str,
    ) -> None:
        self._Captured = f
        self.reverse_id = reverse_id
        self.name = name
        self.description = description

    def __call__(self, *args: Params.args, **kwds: Params.kwargs) -> R:
        return self._Captured(*args, **kwds)




import asyncio
from typing import Callable, Generic, Iterable, Optional, TypeVar, Union

import pandas as pd

from ..types import FlowOutput
from .base import BaseTask

State = TypeVar("State")
PrepareFunc = TypeVar("PrepareFunc", bound=Callable)
RunFunc = TypeVar("RunFunc", bound=Callable)
StopFunc = TypeVar("StopFunc", bound=Callable)
ResultsFunc = TypeVar("ResultsFunc", bound=Callable)
ResultsFunc_ = Callable[
    ['InterpretedTask', FlowOutput],
    Union[Iterable[pd.DataFrame], pd.DataFrame, None]
]


class InterpretedTask(Generic[State], BaseTask):
    def __init__(
        self,
        prepare: PrepareFunc | Callable[['InterpretedTask'], None],
        run: RunFunc | Callable[['InterpretedTask'], None],
        stop: StopFunc | Callable[['InterpretedTask'], None],
        results: ResultsFunc | ResultsFunc_,
        state: State,
        returned: FlowOutput = None,
    ) -> None:
        self.__prepare = prepare
        self.__run = run
        self.__stop = stop
        self.__results = results
        self.__state = state
        self.__returned = returned

    async def async_prepare(
        self,
        callback: Optional[Callable[[], None]] = None,
        *args,
        **kwargs
    ) -> None:
        """Prepare the task for execution.

        Args:
            callback (function, optional): Callback to be run after
                A callback to be run after the task is prepared. A
                function that takes no arguments and returns nothing.
            *args (Any, optional): Positional arguments to be passed to task
                Positional arguments to be passed to the prepare function
                of the produced task. You should check the documentation
                of an interpreter that you are using to see what arguments
                the task expects.
            **kwargs (Any, optional): Keyword arguments to be passed to task
                Keyword arguments to be passed to the prepare function
                of the produced task. You should check the documentation
                of an interpreter that you are using to see what arguments
                the task expects.
        """
        # Run prepare asynchronously and then
        # callback in provided
        await asyncio.sleep(0)
        self.__prepare(self, *args, **kwargs)
        if callback:
            callback()

    async def async_run(
        self,
        callback: Optional[Callable[[], None]] = None,
        *args,
        **kwargs
    ) -> None:
        """Run the task.

        Args:
            callback (function, optional): Callback to be run after
                A callback to be run after the task is prepared. A
                function that takes no arguments and returns nothing.
            *args (Any, optional): Positional arguments to be passed to task
                Positional arguments to be passed to the run function
                of the produced task. You should check the documentation
                of an interpreter that you are using to see what arguments
                the task expects.
            **kwargs (Any, optional): Keyword arguments to be passed to task
                Keyword arguments to be passed to the run function
                of the produced task. You should check the documentation
                of an interpreter that you are using to see what arguments
                the task expects.
        """
        # Run run asynchronously and then
        # callback in provided
        await asyncio.sleep(0)
        self.__run(self, *args, **kwargs)
        if callback:
            callback()

    async def async_stop(
        self,
        callback: Optional[Callable[[], None]] = None,
        *args,
        **kwargs
    ) -> None:
        """Stop the task.

        Args:
            callback (function, optional): Callback to be run after
                A callback to be run after the task is prepared. A
                function that takes no arguments and returns nothing.
            *args (Any, optional): Positional arguments to be passed to task
                Positional arguments to be passed to the stop function
                of the produced task. You should check the documentation
                of an interpreter that you are using to see what arguments
                the task expects.
            **kwargs (Any, optional): Keyword arguments to be passed to task
                Keyword arguments to be passed to the stop function
                of the produced task. You should check the documentation
                of an interpreter that you are using to see what arguments
                the task expects.
        """
        # Run stop asynchronously and then
        # callback in provided
        await asyncio.sleep(0)
        self.__stop(self, *args, **kwargs)
        if callback:
            callback()

    async def async_results(
        self,
        callback: Optional[Callable[[pd.DataFrame | Iterable[pd.DataFrame]], None]] = None,  # noqa: E501
    ) -> None:
        """Get results of the task.

        Args:
            callback (function, optional): Callback to be run after
                A callback to be run after the task is prepared. A
                function that takes no arguments and returns nothing.

                The argument of the callback is a pandas DataFrame
                or an iterable of pandas DataFrames.
        """
        # Run results asynchronously and then
        # callback in provided
        await asyncio.sleep(0)
        results = self.__results(self, self.__returned)
        if callback:
            callback(results)

    def prepare(self, *args, **kwargs) -> None:
        self.__prepare(self, *args, **kwargs)

    def run(self, *args, **kwargs) -> None:
        self.__run(self, *args, **kwargs)

    def stop(self, *args, **kwargs) -> None:
        self.__stop(self, *args, **kwargs)

    def results(self):  # noqa: ANN201
        return self.__results(self, self.__returned)

    def commit_returned(self, returned: FlowOutput):  # noqa: ANN201
        self.__returned = returned

    def __call__(self) -> Union[Iterable[pd.DataFrame], pd.DataFrame, None]:
        self.prepare()
        self.run()
        self.stop()
        return self.results()

    @property
    def state(self) -> State:
        return self.__state



from .._autoflow.tracer import traced, tracedLike
from .._autoflow.tree import ExecutionTree
from ..models.nodes.base import BaseNode
from ..models.nodes.tree import TreeNode


def unwrap_tree(
    tree: ExecutionTree[traced[BaseNode]]
) -> ExecutionTree[traced[BaseNode]]:
    """Merges all subtrees into one tree

    Args
        tree (ExecutionTree): Tree of any depth

    Returns:
        ExecutionTree: Unwrapped tree
    """
    edges = []
    unwraped = {}
    for edge in tree.traverse():
        if isinstance(edge[1].owner, TreeNode):
            if not unwraped.get(edge[1].owner.uuid, False):
                edges.extend(unwrap_tree(edge[1].owner.tree).tree)
                unwraped[edge[1].owner.uuid] = True
            for bridge in edge[2][1]:
                edges.append(
                    (edge[0], bridge[0], bridge[1])
                )
        elif isinstance(edge[0].owner, TreeNode):
            edges.append(
                (tracedLike(edge[0].owner.underlying_node), edge[1], edge[2]),)
        else:
            edges.append(edge)
    return ExecutionTree(edges)



import json
import os
import re
import warnings
from collections import defaultdict
from enum import Enum
from typing import Any, Iterable, Optional
from uuid import uuid4

import pandas as pd
import randomname as rn
from malevich_space.ops.component_manager import ComponentManager
from malevich_space.ops.space import SpaceOps
from malevich_space.schema import VersionMode
from malevich_space.schema.cfg import CfgSchema
from malevich_space.schema.collection_alias import CollectionAliasSchema
from malevich_space.schema.component import ComponentSchema
from malevich_space.schema.flow import (
    FlowSchema,
    InFlowAppSchema,
    InFlowComponentSchema,
    InFlowDependency,
    OpSchema,
    Terminal,
)
from malevich_space.schema.host import LoadedHostSchema
from malevich_space.schema.schema import SchemaMetadata

from malevich.models.nodes.tree import TreeNode

from .._autoflow import tracer as gn
from .._autoflow.tracer import traced
from .._utility.cache.manager import CacheManager
from .._utility.registry import Registry
from .._utility.space.space import resolve_setup
from ..interpreter.abstract import Interpreter
from ..manifest import ManifestManager
from ..models.exceptions import InterpretationError
from ..models.nodes.base import BaseNode
from ..models.nodes.collection import CollectionNode
from ..models.nodes.operation import OperationNode
from ..models.task.interpreted import InterpretedTask
from ..models.types import TracedNode

manf = ManifestManager()
reg = Registry()
cache = CacheManager()


class NodeType(Enum):
    COLLECTION = 'collection'
    OPERATION = 'operation'


class SpaceInterpreterState:
    """State of the Space interpreter."""

    def __init__(self) -> None:

        # A manager to operate with components
        self.component_manager: ComponentManager = None
        # Flow to be interpreted
        self.flow: FlowSchema = FlowSchema()
        # Space Operations (just for convenience, same as in component_manager)
        self.space: SpaceOps = None
        # Host to run the task
        self.host: LoadedHostSchema = None
        # In flow components``
        self.components: dict[str,
                              InFlowComponentSchema | ComponentSchema] = {}
        # In flow components aliases
        self.components_alias: dict[str, str] = {}
        # In flow components configs
        self.components_config: dict[str, dict[str, str]] = {}
        # A mappping from node uuid to node type (collection or operation)
        self.node_to_operation: dict[str, str] = {}
        # A mapping from node uuid to node type (collection or operation)
        self.node_type: dict[str, NodeType] = {}
        # A mapping from node uuid to selected operation in the operation component
        self.selected_operation: dict[str, OpSchema] = {}
        # A mapping from collection uid to CA uid for override (if new data is provided)
        self.collection_overrides: dict[str, str] = {}
        # A mapping from node uuid to dependencies
        self.dependencies: dict[str,
                                list[InFlowDependency]] = defaultdict(list)
        # Unique id for the interpretation
        self.interpretation_id: str = uuid4().hex  # as in API interpreter xD
        # A dictionary for storing auxiliary information
        # task_id, flow_id, etc.
        self.aux: dict[str, Any]
        self.children_states: dict[str, 'SpaceInterpreterState'] = {}

    def copy(self) -> 'SpaceInterpreterState':
        state = SpaceInterpreterState()
        state.component_manager = self.component_manager
        state.flow = self.flow
        state.components = self.components
        state.components_alias = self.components_alias
        state.components_config = self.components_config
        state.dependencies = self.dependencies
        state.aux = self.aux
        state.space = self.space
        state.node_to_operation = self.node_to_operation
        state.node_type = self.node_type
        state.selected_operation = self.selected_operation
        state.interpretation_id = self.interpretation_id
        state.collection_overrides = self.collection_overrides
        state.host = self.host
        state.children_states = self.children_states
        return state


class SpaceInterpreter(Interpreter[SpaceInterpreterState, FlowSchema]):
    supports_subtrees = True

    def prettify_collection_id(
        self,
        collection_id: str
    ) -> str:
        # Replace spaces with dashes
        _s = re.sub(r'\s+', '-', collection_id)
        # Lowercase
        _s = _s.lower()
        # Replace underscores with dashes
        return _s

    def prettify_collection_name(
        self,
        collection_name: str
    ) -> str:
        # Replace multiple spaces with one
        _s = re.sub(r'\s+', ' ', collection_name)
        # Replace dashes and underscores with spaces
        _s = re.sub(r'-|_', ' ', _s)
        # Title case
        return _s.title()

    def prettify_schema_id(self, schema_name: str) -> str:
        _s = re.sub(r'[\s_-]+', ' ', schema_name)
        return _s.replace(' ', '').lower()

    def prettify_component_name(self, component_name: str) -> str:
        _s = re.sub(r'[\s_-]+', ' ', component_name)
        return _s.title()

    def prettify_config_name(self, component_name: str) -> str:
        _s = re.sub(r'[\s_-]+', ' ', component_name)
        return 'Meta Config for ' + _s.title()

    def prettify_config_id(
        self,
        component_name: str,
        rand: Optional[str] = None  # Random string
    ) -> str:
        rand = rand or uuid4().hex[:8]
        __b = 'meta-config-for-' + re.sub(r'[\s|_|-]+', ' ', component_name)
        if rand:
            __b += '-' + rand
        return __b

    def path_for_collection(self, collection_id: str) -> str:
        """Returns a path for the collection in the cache."""
        return os.path.join('space', 'collections', collection_id)

    def update_state(self, state: SpaceInterpreterState = None) -> None:
        """
        The state contains pydantic models, which are not deepcopyable.
        So, this method is used to update the state using custom `copy` method.
        """
        if state:
            self._state = state.copy()
        return state

    @property
    def state(self) -> SpaceInterpreterState:
        return self._state.copy()

    def __init__(
        self,
        name: Optional[str] = None,
        reverse_id: Optional[str] = None,
        description: Optional[str] = None,
    ) -> None:
        """Space Interpreter

        Args:
            name (str): Name of the flow.
            reverse_id (str): Reverse id of the flow.
        """
        super().__init__()
        self._state = SpaceInterpreterState()

        try:
            setup = resolve_setup(manf.query("space", resolve_secrets=True))
        except Exception as e:
            raise InterpretationError(
                "Failed to resolve space setup. "
                "Please check your manifest file.",
                self, self._state
            ) from e

        space = SpaceOps(setup)

        try:  # Local patch for space
            host_ = space.get_my_hosts(url=setup.host.conn_url)[0]
        except Exception:
            host_ = None

        if not host_:
            try:
                host_ = space.create_host(
                    alias=setup.host.alias or '',
                    conn_url=setup.host.conn_url
                )
            except KeyError as ke:
                if 'details' in ke.args[0]:
                    # Local patch for space
                    pass

        self._state.component_manager = ComponentManager(
            host=host_,
            space=space,
            comp_dir=os.path.expanduser('~/.malevich/cache')
        )

        self._state.host = host_
        self._state.space = space

        self._state.aux = {
            'name': name,
            'reverse_id': reverse_id,
            'description': description,
        }

        self.update_state()

    def upload_collection(
        self,
        state: SpaceInterpreterState,
        node: CollectionNode
    ) -> str:
        """Uploads collection (and underlying scheme) to the space."""
        if not node.scheme:
            # Actually, I infer scheme from the collection
            # RFC: Should I allow to upload collection without scheme?
            raise InterpretationError(
                "To be able to use collection, it should have an attached schema",
                self, state
            )

        # Understanding scheme
        if isinstance(node.scheme, SchemaMetadata):
            # If scheme is already parsed
            core_id = node.scheme.core_id

            if not state.space.get_schema(core_id=core_id):
                # HACK: This will fall into another
                # if branch where the scheme will be
                # uploaded
                node.scheme = node.scheme.schema_data

        if isinstance(node.scheme, str):
            # If scheme is a string, then it should firstly
            # be uploaded
            core_id = self.prettify_schema_id(
                node.collection.collection_id
            ),

            schema_uid = state.space.create_scheme(
                # $core_id: String!, $raw: String!, $name: String
                core_id=core_id,
                raw=node.scheme,
                name=self.prettify_schema_id(
                    node.collection.collection_id
                )
            )
            # RFC: Should I allow to upload collection without scheme?
            assert schema_uid, "Failed to upload the schema"

        # Uploading collection
        alias = self.prettify_collection_name(
            node.collection.collection_id
        )

        # RFC!: Is it correct way to upload collection?
        uid = state.space.create_collection(
            host_id=state.host.uid,
            core_id=node.collection.collection_id,
            core_alias=alias,
            schema_core_id=core_id,
            docs=[
                row.to_json()
                for _, row in node.collection.collection_data.iterrows()
            ]
        )

        return uid

    def interpret(self, node: TreeNode) -> InterpretedTask:
        self._state.aux['name'] = self._state.aux['name'] or node.name
        self._state.aux['reverse_id'] = self._state.aux['reverse_id'] or node.reverse_id
        self._state.aux['description'] = self._state.aux['description'] or node.description  # noqa: E501
        return super().interpret(node)

    def before_interpret(self, state) -> SpaceInterpreterState:
        state.flow = FlowSchema()
        return state

    def create_node(  # Let's go!
        self, state: SpaceInterpreterState, node: gn.traced[BaseNode]
    ) -> SpaceInterpreterState:
        """Wraps node in the execution graph with a Malevich Space component schema.

        Args:
            state (SpaceInterpreterState): Interpreter state.
            node (gn.traced[BaseNode]): Node to wrap.

        Returns:
            SpaceInterpreterState: Interpreter state.
        """
        # TODO: The branch is too long, should refactor it

        if isinstance(node.owner, CollectionNode):  # If the node is a collection
            state.node_type[node.owner.uuid] = NodeType.COLLECTION
            core_schema_id = self.prettify_schema_id(
                node.owner.collection.collection_id
            )
            # Create a schema for the collection (only name and reverse id)
            # Maybe schema is already created
            if (schema := state.component_manager.space.get_schema(core_id=core_schema_id)) is None:  # noqa: E501
                # Create the schema if not created
                schema_id = state.component_manager.space.create_scheme(
                    # core.create_scheme(...)
                    core_id=core_schema_id,
                    # This is generated with wrapped
                    # class
                    # See malevich._meta.collection.collection()
                    raw=node.owner.scheme
                )

                # And another one!
                schema = state.component_manager.space.get_schema(
                    uid=schema_id, core_id=core_schema_id
                )

            # To upload the collection, it is required to
            # save the collection data in a csv file
            cache_collection_path = cache.make_path_in_cache(
                # Utilizing `~/.malevich/cache`
                self.path_for_collection(node.owner.collection.collection_id)
            )
            # Save the collection data in the csv file
            if not node.owner.collection.collection_data.empty:
                path = self.path_for_collection(
                    node.owner.collection.collection_id)
                node.owner.collection.collection_data.to_csv(
                    cache_collection_path,
                    index=False
                )
            else:
                path = None

            # Try to get the collection component from the space
            component = state.space.get_parsed_component_by_reverse_id(
                # Using the collection id as the reverse id
                reverse_id=node.owner.collection.collection_id
            )

            # If the component is not found (not uploaded yet)
            if component is None:
                # Wrap the collection in a component schema
                name = self.prettify_collection_name(
                    node.owner.collection.collection_id
                )
                comp = ComponentSchema(
                    name=name,
                    description=f"Meta collection {name}",
                    reverse_id=node.owner.collection.collection_id,
                    collection=CollectionAliasSchema(
                        core_alias=node.owner.collection.collection_id,
                        schema_core_id=schema.core_id,
                        path=path
                    )
                )
            else:
                # Collection already exists, so two
                # cases:
                # 1. Overriding with new data
                # 2. Using as it is
                coll_id = self.prettify_collection_id(
                    node.owner.collection.collection_id
                )
                if not node.owner.collection.collection_data.empty:
                    # Has some data, so override
                    uid = state.space.create_collection(
                        host_id=state.host.uid,
                        # core_id=f'override-{coll_id}-{state.interpretation_id}',
                        core_alias=f'override-{coll_id}-{state.interpretation_id}',
                        schema_core_id=schema.core_id,
                        docs=[
                            row.to_json()
                            for _, row in node.owner.collection.collection_data.iterrows()  # noqa: E501
                        ]
                    )
                    state.collection_overrides[component.collection.uid] = uid

                    comp = ComponentSchema(
                        reverse_id=component.reverse_id,
                        name=component.name,
                    )

        elif isinstance(node.owner, OperationNode):  # If the node is an operation
            state.node_type[node.owner.uuid] = NodeType.OPERATION

            # All the required information to interpret the node
            # should be in the registry
            extra = reg.get(node.owner.operation_id)

            # If no extra information is found, it means
            # local information is too old
            # (but it is user fault, not mine)
            if 'reverse_id' not in extra:
                raise InterpretationError(
                    "Trying to interpret an operation that is not installed "
                    "properly or extremely outdated. Try to reinstall "
                    "the operation. ",
                    self, state
                )

            # Get the component from the space (it should be there)
            component = state.component_manager.space.get_parsed_component_by_reverse_id(  # noqa: E501
                reverse_id=extra['reverse_id']
            )

            if not component:
                # How then you install it? Was it wiped? :(
                raise InterpretationError(
                    "Trying to interpret an operation that is not installed "
                    "properly or extremely outdated. Try to reinstall "
                    "the operation with\n\t"
                    f"`malevich install {extra['reverse_id']} --using space`",
                    self, state
                )

            state.components_config[node.owner.uuid] = node.owner.config
            comp = ComponentSchema(
                name=component.name,
                reverse_id=component.reverse_id,
                app=component.app
            )
            state.node_to_operation[node.owner.uuid] = node.owner.operation_id

        elif isinstance(node.owner, TreeNode):
            if not state.children_states.get(node.owner.uuid, None):
                child_interpreter = SpaceInterpreter(
                    name=node.owner.name,
                    reverse_id=node.owner.reverse_id
                )

                child_interpreter.interpret(node.owner)
                child_state: SpaceInterpreterState = child_interpreter.state

                comp = ComponentSchema(
                    name=node.owner.name,
                    reverse_id=node.owner.reverse_id,
                    flow=child_state.flow,
                )

                state.children_states[node.owner.uuid] = child_state
            else:
                child_state = state.children_states[node.owner.uuid]
                comp = ComponentSchema(
                    name=node.owner.name,
                    reverse_id=node.owner.reverse_id,
                    flow=child_state.flow,
                )

        if not comp:
            raise InterpretationError(
                "Failed to interpret the node. This is a bug, please report it.",
            )

        state.components[node.owner.uuid] = comp
        # FIXME: state.components_alias are not neccessary then...
        state.components_alias[node.owner.uuid] = node.owner.alias
        return state

    def create_dependency(
        self,
        state: SpaceInterpreterState,
        caller: TracedNode,
        callee: traced[OperationNode],
        # NOTE: This is an argument name. The fact
        # that it is not used means Space does not accept
        # multiple collections as input
        link: tuple[int, list[tuple[BaseNode, str]]] | str
    ) -> SpaceInterpreterState:
        """Creates a dependency between two nodes."""

        # Case Collection / App -> Flow

        if isinstance(callee.owner, TreeNode) and not isinstance(caller.owner, TreeNode):  # noqa: E501
            child = state.children_states[callee.owner.uuid]
            caller_alias = state.components_alias[caller.owner.uuid]
            inter_flow_map = {}

            bridges = [l_ for l_ in link[1]]

            for to, _ in bridges:
                inter_flow_map[caller_alias] = child.components_alias[to.owner.uuid]

            dependency = InFlowDependency(
                from_op_id=(
                    # provided by space installer
                    caller.owner.operation_id
                    if isinstance(caller.owner, OperationNode)
                    else None
                ),
                # ??
                to_op_id=callee.owner.underlying_node.operation_id if isinstance(
                    callee.owner.underlying_node, OperationNode) else None,
                alias=state.components_alias[caller.owner.uuid],
                terminals=[
                    Terminal(
                        src=x,
                        target=y
                    ) for x, y in inter_flow_map.items()
                ]
            )
        elif isinstance(caller.owner, TreeNode) and not isinstance(callee.owner, TreeNode):  # noqa: E501
            child = state.children_states[caller.owner.uuid]
            callee_alias = state.components_alias[callee.owner.uuid]
            op: OperationNode = caller.owner.underlying_node
            inter_flow_map = {
                child.components_alias[op.uuid]: callee_alias
            }

            dependency = InFlowDependency(
                from_op_id=op.operation_id,
                to_op_id=(
                    # provided by space installer
                    callee.owner.operation_id
                    if isinstance(callee.owner, OperationNode)
                    else None
                ),
                alias=state.components_alias[caller.owner.uuid],
                terminals=[
                    Terminal(
                        src=x,
                        target=y
                    ) for x, y in inter_flow_map.items()
                ]
            )
        else:
            dependency = InFlowDependency(
                from_op_id=(
                    # provided by space installer
                    caller.owner.operation_id
                    if isinstance(caller.owner, OperationNode)
                    else None
                ),
                to_op_id=callee.owner.operation_id,
                alias=state.components_alias[caller.owner.uuid],
            )

        state.dependencies[callee.owner.uuid].append(dependency)

        return state

    def after_interpret(self, state: SpaceInterpreterState) -> SpaceInterpreterState:
        """Finishes the interpretation by adding components to the flow."""
        for uid, component in state.components.items():
            try:
                # I don't know why, but sometimes
                # it fails, so try/exc here
                loaded_component = state.component_manager.component(
                    component,
                    VersionMode.DEFAULT
                )
            except Exception as e:
                # If it fails, try to get the component
                # by reverse id
                loaded_component = state.component_manager.space.get_component_by_reverse_id(  # noqa: E501
                    component.reverse_id
                )

                # No components?(
                # Nothing to do here
                if loaded_component is None:
                    raise InterpretationError(
                        f"Failed to interpret the flow: component {component.name} "
                        "failed to load. ",
                        self, state
                    ) from e

            # If the component is an operation
            # and has config
            if uid in state.components_config:
                # RFC: Is it correct way to update config?
                space_config = CfgSchema(
                    readable_name=self.prettify_config_name(
                        component.reverse_id),
                    cfg_json=state.components_config.get(uid),
                    core_name=self.prettify_config_id(component.reverse_id),
                )
            else:
                space_config = None

            if uid in state.node_to_operation:
                extra = reg.get(state.node_to_operation[uid], {})
            else:
                extra = {}

            # Add the component to the flow
            state.flow.components = [
                *state.flow.components,
                InFlowComponentSchema(
                    reverse_id=component.reverse_id,
                    alias=state.components_alias[uid],
                    depends={
                        dep.alias: dep
                        for dep in state.dependencies[uid]
                    },
                    app=InFlowAppSchema(
                        active_op=[
                            OpSchema(
                                core_id=extra['processor_name'],
                                type='processor'
                            )
                        ]
                    ) if component.app is not None else None,
                    active_cfg=space_config,
                ),
            ]

        return state

    def get_task(
        self, state: SpaceInterpreterState
    ) -> InterpretedTask[SpaceInterpreterState]:
        component = state.component_manager.component(
            ComponentSchema(
                name=state.aux['name'],
                description=f"Meta flow {state.aux['name']}!",
                reverse_id=state.aux['reverse_id'],
                flow=state.flow,
                # version=VersionSchema(
                #     readable_name='0.0.0',
                # )
            ),
            VersionMode.PATCH
        )
        save_flow_path = cache.make_path_in_cache(
            os.path.join(
                'flows', f'{state.aux["reverse_id"]}_{state.interpretation_id}.json'
            )
        )

        with open(save_flow_path, 'w+') as f:
            json.dump(state.flow.model_dump(), f, indent=2)

        def prepare(
            task: InterpretedTask[SpaceInterpreterState],
            *args,
            **kwargs
        ) -> None:
            task.state.aux['flow_id'] = component.flow.uid

            task_id = state.space.build_task(
                flow_id=component.flow.uid,
                host_id=state.component_manager.host.uid
            )

            state.aux['task_id'] = task_id[0]

            state.space.boot_task(
                task_id=state.aux['task_id'],
            )

        def run(
            task: InterpretedTask[SpaceInterpreterState],
            *args,
            **kwargs
        ) -> None:
            if not state.aux.get('task_id'):
                raise Exception(
                    "Attempt to run a task which is not prepared. "
                    "Please prepare the task first."
                )

            start_schema = state.space.get_task_start_schema(
                state.aux['task_id'],
            )

            in_flow_ca = {}
            overrides = []
            for sch in start_schema:
                try:
                    in_flow_ca = state.space.get_ca_in_flow(
                        flow_id=task.state.aux['flow_id'],
                        in_flow_id=sch.in_flow_id
                    )

                    overrides.append({
                        "inFlowCompUid": sch.in_flow_id,
                        "caUid": task.state.collection_overrides[in_flow_ca],
                        "caAlias": sch.injected_alias
                    })
                except Exception:
                    # TODO fix!
                    continue

            state.space.run_task(
                task_id=state.aux['task_id'],
                ca_override=overrides
            )

        def results(
            task: InterpretedTask[SpaceInterpreterState],
            returned: Iterable[traced[BaseNode]] | traced[BaseNode] | None,
            *args,
            **kwargs
        ) -> Iterable[pd.DataFrame] | pd.DataFrame:
            warnings.warn("Currently Space interpreter does not return results. Please "
                          "use UI interface to work with flow")
            return []

        return InterpretedTask(
            prepare=prepare,
            run=run,
            # TODO: Stopping task on space
            stop=lambda *args, **kwargs: None,
            results=results,
            state=state,
        )



from typing import Annotated, Optional

import rich
import typer
from rich.progress import Progress, SpinnerColumn, TextColumn

from malevich.install.space import SpaceInstaller

from ..constants import (
    DEFAULT_CORE_HOST,
    IMAGE_BASE,
    USE_HELP,
    USE_IMAGE_HELP,
)
from ..install.image import ImageInstaller
from ..manifest import ManifestManager

use = typer.Typer(help=USE_HELP, rich_markup_mode="rich")


def _install_from_image(
    package_name: str,
    image_ref: Optional[str] = None,
    image_auth_user: Optional[str] = None,
    image_auth_password: Optional[str] = None,
    core_host: str = DEFAULT_CORE_HOST,
    core_user: Optional[str] = None,
    core_token: Optional[str] = None,
    progress: Progress = None,

) -> None:
    if image_ref is None:
        return _install_from_image(
            package_name=package_name,
            image_ref=IMAGE_BASE.format(app=package_name),
            image_auth_user=image_auth_user,
            image_auth_password=image_auth_password,
            core_host=core_host or DEFAULT_CORE_HOST,
            core_user=core_user,
            core_token=core_token,
            progress=Progress(
                SpinnerColumn(), TextColumn("{task.description}")),
        )
    installer = ImageInstaller()
    if progress:
        task_ = progress.add_task(
            f"Attempting to install [b blue]{package_name}[/b blue] with"
            " [i yellow]image[/i yellow] installer",
            total=1,
        )

    try:
        manifest_entry = installer.install(
            package_name=package_name,
            image_ref=image_ref,
            image_auth=(image_auth_user, image_auth_password),
            core_host=core_host,
            core_auth=(
                core_user, core_token) if core_user and core_token else None,
        )

        manifest_manager = ManifestManager()
        if entry := manifest_manager.query("dependencies", package_name):
            if entry.get("installer") != "image":
                raise Exception(
                    """Package {package_name} already exists with different installer.
                    Possible solutions:
                        - Remove package from manifest and try again. Use [b]malevich remove[/b] command
                        - Use [b]malevich restore[/b] command to restore package with different installer
                        - Use [b]malevich install[/b] to install the package with all possible installers"""  # noqa: E501
                )

            manifest_manager.put(
                "dependencies",
                package_name,
                value=manifest_entry.model_dump(),
            )
            if progress:
                progress.update(
                    task_,
                    description="\n[green]✔[/green] Package "
                    f"[blue]({package_name})[/blue] "
                    "[yellow]updated[/yellow] successfully",
                    completed=1,
                )
        else:
            if progress:
                progress.update(
                    task_,
                    description="\n[green]✔[/green] Package "
                    f"[blue]({package_name})[/blue] "
                    "installed successfully",
                    completed=1,
                )
            manifest_manager.put(
                "dependencies",
                value={f"{package_name}": manifest_entry.model_dump()},
                append=True,
            )

        manifest_manager.save()
        return True
    except Exception as err:
        if progress:
            progress.update(
                task_,
                description="\n[red]✘[/red] Failed with "
                "[yellow]image[/yellow] installer "
            )
        raise err


@use.command("image", help=USE_IMAGE_HELP)
def install_from_image(
    package_name: Annotated[
        str,
        typer.Argument(help="Package name")
    ],
    image_ref: Annotated[
        Optional[str],
        typer.Argument(
            help="Image reference in format <registry>/<image>:<tag>"
        )
    ] = None,
    image_auth_user: Annotated[
        str,
        typer.Argument(
            help="Image registry auth user"
        )
    ] = "",
    image_auth_password: Annotated[
        str,
        typer.Argument(
            help="Image registry auth password"
        )
    ] = "",
    core_host: Annotated[
        str,
        typer.Option(
            help="Malevich Core hostname"
        )
    ] = DEFAULT_CORE_HOST,
    core_user: Annotated[
        Optional[str],
        typer.Option(
            help="Malevich Core user"
        )
    ] = None,
    core_token: Annotated[
        Optional[str],
        typer.Option(
            help="Malevich Core token"
        )
    ] = None,
) -> None:
    try:
        if image_ref is None:
            return _install_from_image(
                package_name=package_name,
                image_ref=IMAGE_BASE.format(app=package_name),
                image_auth_user=image_auth_user,
                image_auth_password=image_auth_password,
                core_host=core_host or DEFAULT_CORE_HOST,
                core_user=core_user,
                core_token=core_token,
                progress=Progress(
                    SpinnerColumn(), TextColumn("{task.description}")),
            )
        else:
            return _install_from_image(
                package_name=package_name,
                image_ref=image_ref,
                image_auth_user=image_auth_user,
                image_auth_password=image_auth_password,
                core_host=core_host or DEFAULT_CORE_HOST,
                core_user=core_user,
                core_token=core_token,
                progress=Progress(
                    SpinnerColumn(), TextColumn("{task.description}")),
            )
    except Exception as err:
        rich.print("\n\n[red]Installation failled[/red]")
        rich.print(err)
        exit(-1)


def _install_from_space(
    package_name: str,
    reverse_id: str,
    branch: Optional[str] = None,
    version: Optional[str] = None,
    progress: Progress = None,
) -> None:
    installer = SpaceInstaller()
    if progress:
        task_ = progress.add_task(
            f"Attempting to install [b blue]{package_name}[/b blue] with"
            " [i yellow]space[/i yellow] installer",
            total=1,
        )

    try:
        manifest_entry = installer.install(
            package_name=package_name,
            reverse_id=reverse_id,
            branch=branch,
            version=version,
        )
        manifest_manager = ManifestManager()
        if entry := manifest_manager.query("dependencies", package_name):
            if entry.get("installer") != "space":
                raise Exception(
                    f"Package {package_name} already exists with different installer."
                    "\nPossible solutions:"
                    "\n\t- Remove package from manifest and try again. Use `malevich remove` command"  # noqa: E501
                    "\n\t- Use `malevich restore` command to restore package with different installer"  # noqa: E501
                    "\n\t- Use `malevich install` to install the package with all possible installers"  # noqa: E501
                )

            manifest_manager.put(
                "dependencies",
                package_name,
                value=manifest_entry.model_dump(),
            )
            if progress:
                progress.update(
                    task_,
                    description="\n[green]✔[/green] Package "
                    f"[blue]({package_name})[/blue] "
                    "[yellow]updated[/yellow] successfully",
                    completed=1,
                )
        else:
            if progress:
                progress.update(
                    task_,
                    description="\n[green]✔[/green] Package "
                    f"[blue]({package_name})[/blue] "
                    "installed successfully",
                    completed=1,
                )
            manifest_manager.put(
                "dependencies",
                value={f"{package_name}": manifest_entry.model_dump()},
                append=True,
            )

        manifest_manager.save()
        return True
    except Exception as err:
        if progress:
            progress.update(
                task_,
                description="\n[red]✘[/red] Failed with "
                "[yellow]space[/yellow] installer "
            )
        raise err


@use.command("space", help=USE_IMAGE_HELP)
def install_from_space(
    package_name: Annotated[
        str,
        typer.Argument(help="Package name")
    ],
    reverse_id: Annotated[
        str,
        typer.Argument(help="Reverse id of the component")
    ],
    branch: Annotated[
        Optional[str],
        typer.Argument(help="Branch of the component")
    ] = None,
    version: Annotated[
        Optional[str],
        typer.Argument(help="Version of the component")
    ] = None,
) -> None:
    try:
        return _install_from_space(
            package_name=package_name,
            reverse_id=reverse_id,
            branch=branch,
            version=version,
            progress=Progress(
                SpinnerColumn(), TextColumn("{task.description}")),
        )
    except Exception as err:
        rich.print("\n\n[red]Installation failled[/red]")
        rich.print(err)
        exit(-1)



import pandas as pd


def pd_to_json_schema(df: pd.DataFrame, format=False) -> dict:
    mappings = {
        'integer': 'number',
        'float': 'number',
        'boolean': 'boolean',
        'string': 'string',
        'object': 'object',
        'datetime': 'string',
        'date': 'string',
        'datetime64[ns]': 'string',
        'category': 'string',
        'timedelta[ns]': 'string',
    }

    schema = {
        'type': 'object',
        'properties': {},
        'required': [],
    }

    for col in df.columns:
        col_type = df[col].dtype.name
        schema['properties'][col] = {
            'type': mappings.get(col_type, 'string'),
        }
        if format:
            if col_type == 'object':
                schema['properties'][col]['format'] = 'date-time'
            if col_type == 'datetime64[ns]':
                schema['properties'][col]['format'] = 'date-time'
            if col_type == 'date':
                schema['properties'][col]['format'] = 'date'
            if col_type == 'boolean':
                schema['properties'][col]['format'] = 'boolean'
            if col_type == 'integer':
                schema['properties'][col]['format'] = 'int64'
            if col_type == 'float':
                schema['properties'][col]['format'] = 'float64'

    return schema



import enum
import os
import tempfile

import requests
from git import Repo
from github import Github

from .base import CIOps


class DockerRegistry(enum.Enum):
    OTHER = "other"
    PUBLIC_AWS_ECR = "ecr"
    PRIVATE_AWS_ECR = "ecr-private"
    YANDEX = "yandex"
    GCR = "gcr"


class GithubCIOps(CIOps):
    def __has_malevich_actions(self, repo_dir: str) -> bool:
        return os.path.exists(
            os.path.join(repo_dir, ".github", "workflows", "malevich-ci.yml")
        )

    def __create_action(self, repo_dir: str, branch: str) -> str:
        os.makedirs(os.path.join(repo_dir, ".github",
                    "workflows"), exist_ok=True)

        action_file_path = os.path.join(
            repo_dir, ".github", "workflows", f"malevich-ci_{branch}.yml"
        )

        template_file_path = os.path.join(
            os.path.dirname(__file__), 'templates', 'malevich-ci.yml'
        )

        with open(action_file_path, "w") as action, \
                open(template_file_path) as template:

            action.write(template.read().format(
                BRANCH_NAME=branch
            ))

        return action_file_path

    def __create_manual_action(self, repo_dir: str, branch: str) -> str:
        os.makedirs(os.path.join(repo_dir, ".github",
                    "workflows"), exist_ok=True)

        action_file_path = os.path.join(
            repo_dir, ".github", "workflows", f"malevich-ci-manual_{branch}.yml"
        )

        template_file_path = os.path.join(
            os.path.dirname(__file__), 'templates', 'malevich-ci-manual.yml'
        )

        with open(action_file_path, "w") as action, \
                open(template_file_path) as template:

            action.write(template.read().format(
                BRANCH_NAME=branch
            ))

        return action_file_path

    def _log(self, verbose: bool, msg: str) -> None:
        if verbose:
            print(msg)

    def setup_ci(
        self,
        token: str,
        repository: str,
        space_user: str,
        space_token: str,
        space_url: str,
        branch: str,
        registry_url: str,
        registry_id: str,
        image_user: str,
        image_token: str,
        org_id: str,
        verbose: bool = False,
    ) -> None:
        """Setup CI for a Github repository

        Args:
            token (str): Github token
            repository (str): Github repository in the form of `user/repo`
            space_user (str): Malevich Space username
            space_token (str): Malevich Space token (password)
            space_url (str): Malevich Space API URL
            branch (str): Branch to setup CI in
            registry_url (str):
                Docker Image Registry URL, for example `public.ecr.aws` or 'cr.yandex'
            registry_id (str): Docker Registry ID
            registry_type (DockerRegistry): Docker Registry type
            image_user (str): Username to access the Docker Image Registry
            image_token (str): TOKEN to access the Docker Image Registry
            org_id (str): ORGANIZATION_ID
            verbose (bool, optional): Verbose mode. Defaults to False.
        """
        g = Github(token)

        repo = g.get_repo(repository)

        repo.create_environment(branch)
        pub_key = repo.get_public_key()
        repo_id = repo.id

        keys = [
            "SPACE_USERNAME",
            "SPACE_PASSWORD",
            "API_URL",
            "IMAGE_USERNAME",
            "IMAGE_PASSWORD",
            "REGISTRY_URL",
            "REGISTRY_ID",
            "REGISTRY_TYPE",
            "SPACE_ORGANIZATION_ID"
        ]

        values = [
            space_user,
            space_token,
            space_url,
            image_user,
            image_token,
            registry_url,
            registry_id,
            'ecr' if 'ecr' in registry_url else 'other',
            org_id
        ]

        for key, value in zip(keys, values):
            payload = pub_key.encrypt(value)
            url = f'https://api.github.com/repositories/{repo_id}/environments/{branch}/secrets/{key}'
            requests.put(
                url,
                headers={
                    "Accept": "application/vnd.github+json",
                    "Authorization": f"Bearer {token}",
                    "X-GitHub-Api-Version": "2022-11-28"
                },
                json={
                    'encrypted_value': f'{payload}',
                    'key_id': f'{pub_key.key_id}'
                }
            )
            self._log(verbose, f"Updated secret {key}")

        with tempfile.TemporaryDirectory() as tmpdir:
            git_repo = Repo.clone_from(repo.clone_url, tmpdir)
            self._log(verbose, f"Cloned repository {repository} into {tmpdir}")
            head = git_repo.create_head(branch)
            head.checkout()
            self._log(verbose, f"Checkouted to branch {branch}")
            action_file = self.__create_action(tmpdir, branch)
            manual_action_file = self.__create_manual_action(tmpdir, branch)
            self._log(verbose, f"Created action file {action_file}")
            git_repo.index.add([action_file])
            git_repo.index.add([manual_action_file])
            git_repo.index.commit(
                "add: malevich-ci.yml && malevich-ci-manual.yml")
            self._log(verbose,
                      f"Committed actions file {action_file}, {manual_action_file}"
                      )
            origin = git_repo.remote(name="origin")
            self._log(verbose, "Pushing to origin")
            origin.push(git_repo.active_branch)
            self._log(verbose, "Pushed to origin")



import json


def lhash_json(x: object, length=6, exp=32) -> str:
    _s = json.dumps(x, sort_keys=True)
    _i = hash(_s) % (2 ** exp)
    # Map the integer to a string with
    # characters a-z, A-Z, 0-9

    _s = ""
    alph = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
    for _ in range(length):
        _s += alph[_i % len(alph)]
        _i //= len(alph)
    return _s



import re
from typing import Optional

from malevich_space.ops import SpaceOps
from pydantic import BaseModel

from .._utility.space.space import resolve_setup
from ..manifest import ManifestManager
from ..models.manifest import Dependency
from .installer import Installer
from .mimic import mimic_package

manf = ManifestManager()


class Templates:
    """Templates for creating a package from a Docker image"""

    disclaimer = """\"\"\"
THIS FILE IS AUTOGENERATED BY malevich PACKAGE INSTALLER.

THIS FILE CONTAINS VITAL INFORMATION ABOUT THE PACKAGE AND ITS CONTENTS.
DO NOT MODIFY THIS FILE MANUALLY.
\"\"\"
"""

    imports = """
from malevich._autoflow.function import autotrace
from malevich._utility.registry import Registry
from malevich.models.nodes import OperationNode

from pydantic import BaseModel
"""
    registry = """
Registry().register("{operation_id}", {{
    "reverse_id": "{reverse_id}",
    "branch": {branch},
    "version": {version},
    "processor_name": "{name}",
    "processor_id": "{name}",
    "image_ref": {image_ref},
    "image_auth_user": {image_auth_user},
    "image_auth_pass": {image_auth_pass},
}})
"""

    processor = """
@autotrace
def {name}({args}config: dict = {{}}):
    \"\"\"{docs}\"\"\"
    return OperationNode(operation_id="{operation_id}", config=config)
"""


class SpaceDependencyOptions(BaseModel):
    reverse_id: str
    branch: Optional[str] = None
    version: Optional[str] = None
    image_ref: Optional[str] = None
    image_auth_user: Optional[str] = None
    image_auth_pass: Optional[str] = None


class SpaceDependency(Dependency):
    options: SpaceDependencyOptions


class SpaceInstaller(Installer):
    name = 'space'

    def __init__(self) -> None:
        super().__init__()
        try:
            self.__ops = SpaceOps(resolve_setup(
                manf.query('space', resolve_secrets=True)
            ))
        except Exception as e:
            raise Exception(
                "Setup is invalid. Run `malevich space init`") from e

    def install(
        self,
        package_name: str,
        reverse_id: str,
        branch: Optional[str] = None,
        version: Optional[str] = None,
    ) -> SpaceDependency:

        component = self.__ops.get_parsed_component_by_reverse_id(
            reverse_id=reverse_id)
        if component is None:
            raise Exception(f"Component {reverse_id} not found")

        if component.app is None:
            raise Exception(f"Component {reverse_id} is not an app")

        metascript = Templates.disclaimer
        metascript += Templates.imports

        m = ManifestManager()

        if component.app:
            iauth, itoken = m.put_secrets(
                image_auth_user=component.app.container_user,
                image_auth_password=component.app.container_token,
            )
        else:
            iauth, itoken = None, None

        for op in component.app.ops:
            if op.type != "processor":
                continue
            metascript += Templates.registry.format(
                operation_id=op.uid,
                reverse_id=reverse_id,
                branch=str(component.branch.model_dump()),
                version=str(component.version.model_dump()),
                name=op.core_id,
                image_ref=("dependencies", package_name,
                           "options", "image_ref"),
                image_auth_user=(
                    "dependencies",
                    package_name,
                    "options",
                    "image_auth_user",
                ),
                image_auth_pass=(
                    "dependencies",
                    package_name,
                    "options",
                    "image_auth_pass",
                ),
            )

            args_ = []
            for arg_ in op.args:
                if "return" in arg_.arg_name \
                        or (arg_.arg_type and "Context" in arg_.arg_type):
                    continue
                args_.append(arg_.arg_name)

            metascript += Templates.processor.format(
                name=op.core_id,
                args=", ".join(args_) + ", " if args_ else "",
                docs=op.doc,
                operation_id=op.uid
            )

        mimic_package(
            package=re.sub(r"[^a-zA-Z0-9_]", "_", package_name),
            metascript=metascript,
        )

        return SpaceDependency(
            package_id=package_name,
            version=component.version.readable_name,
            installer="space",
            options=SpaceDependencyOptions(
                reverse_id=reverse_id,
                branch=component.branch.uid,
                version=component.version.uid,
                image_ref=component.app.container_ref,
                image_auth_user=iauth,
                image_auth_pass=itoken
            )
        )

    def restore(self, dependency: SpaceDependency) -> None:
        return self.install(
            package_name=dependency.package_id,
            reverse_id=dependency.options.reverse_id,
            branch=dependency.options.branch,
            version=dependency.options.version
        )

    def construct_dependency(self, dependency: dict) -> SpaceDependency:
        return SpaceDependency(**dependency)



from typing import Any, Optional

from malevich_space.schema import SpaceSetup
from pydantic import BaseModel


class Dependency(BaseModel):
    package_id: str
    version: str
    installer: str
    options: Any

    def simple(self) -> dict[str, Any]:
        return {
            **self.options.model_dump()
        }


class Manifest(BaseModel):
    project_id: Optional[str] = None
    version: Optional[str] = None
    space: Optional[SpaceSetup] = None
    dependencies: list[dict[str, Dependency]] = []


class ManifestUpdateEntry(BaseModel):
    package_id: str
    version: str
    installer: str
    options: dict[str, Any]


class Secret(BaseModel):
    secret_key: str
    secret_value: str
    salt: Optional[str] = None
    external: Optional[bool] = False


class Secrets(BaseModel):
    secrets: dict[str, Secret] = {}



import os
import shutil
import sys

from .singleton import SingletonMeta


class PackageManager(metaclass=SingletonMeta):
    builtins = [
        "_autoflow",
        "_meta",
        "_utility",
        "commands",
        "install",
        "interpreter",
        "models",
        "square",
    ]

    def __init__(self) -> None:
        self.__malevich_path = os.path.dirname(
            sys.modules["malevich"].__file__)

    def is_builtin(self, package_name: str) -> bool:
        return package_name in self.builtins

    def get_package_path(self, package_name: str) -> str:
        return os.path.join(self.__malevich_path, package_name)

    def create_stub(self, package_name: str, package_content: str) -> None:
        package_path = self.get_package_path(package_name)
        if os.path.exists(package_path):
            raise Exception(f"Package {package_name} already exists")
        os.mkdir(package_path)
        with open(os.path.join(package_path, "__init__.py"), "w") as f:
            f.write(package_content)

    def remove_stub(self, package_name: str) -> None:
        package_path = self.get_package_path(package_name)
        if not os.path.exists(package_path):
            raise Exception(f"Package {package_name} does not exist")
        shutil.rmtree(package_path)

    def get_package(self, package_name: str) -> None:
        package_path = self.get_package_path(package_name)
        if not os.path.exists(package_path):
            raise Exception(f"Package {package_name} does not exist")
        return package_path

    def get_all_packages(self) -> None:
        return [
            package
            for package in os.listdir(self.__malevich_path)
            if os.path.isdir(os.path.join(self.__malevich_path, package))
            and package not in self.builtins
        ]

    def is_importable(self, package_name: str) -> bool:
        import importlib
        try:
            lib = importlib.import_module(f"malevich.{package_name}")
        except Exception:
            return False
        return lib is not None


package_manager = PackageManager()



import os
from typing import IO

from ..singleton import SingletonMeta


class CacheManager(metaclass=SingletonMeta):
    def __init__(self) -> None:
        self._fs = os.path.expanduser('~/.malevich/cache')

    def get_file(self, path: str, mode: str) -> IO | None:
        return open(self.make_path_in_cache(path), mode)

    def make_path_in_cache(self, base_path: str) -> str:
        os.makedirs(
            os.path.dirname(os.path.join(self._fs, base_path),),
            exist_ok=True
        )
        return os.path.join(self._fs, base_path)
